/* 
 * Copyright (c) 2001-$Date: 2006-05-01 14:41:34 -0700 (Mon, 01 May 2006) $ TIBCO Software Inc. 
 * All rights reserved.
 * For more information, please contact:
 * TIBCO Software Inc., Palo Alto, California, USA
 * 
 * $Id: tibjmsAsyncMsgConsumerRecover.java 21731 2006-05-01 21:41:34Z rselph $
 * 
 */

/*
 * This is a simple sample of a basic asynchronous
 * tibjmsMsgConsumer.
 *
 * This sample subscribes to specified destination and
 * receives and prints all received messages.
 *
 * Notice that the specified destination should exist in your configuration
 * or your topics/queues configuration file should allow
 * creation of the specified destination.
 *
 * If this sample is used to receive messages published by
 * tibjmsMsgProducer sample, it must be started prior
 * to running the tibjmsMsgProducer sample.
 *
 * Usage:  java tibjmsAsyncMsgConsumerRecover [options]
 *
 *    where options are:
 *
 *      -server     Server URL.
 *                  If not specified this sample assumes a
 *                  serverUrl of "tcp://localhost:7222"
 *
 *      -user       User name. Default is null.
 *      -password   User password. Default is null.
 *      -topic      Topic name. Default is "topic.sample"
 *      -queue      Queue name. No default
 *
 *
 */
/*
 * Slightly modified TIBCO example to automatically reconnect to the EMS server and demonstrate session.recover() 
 * and message re-delivery and setting client heartbeat settings
 * 2012-12-10 matti.kaikkonen@teliasonera.com
 * 
 */

import javax.jms.*;

import com.tibco.tibjms.Tibjms;

public class tibjmsAsyncMsgConsumerRecover
       implements ExceptionListener, MessageListener
{
    /*-----------------------------------------------------------------------
     * Parameters
     *----------------------------------------------------------------------*/

    String    serverUrl   = "tcp://localhost:7222,tcp://localhost:7222";
    String    userName    = null;
    String    password    = null;
    String    name        = "topic.sample";
    String    replyName   = "sample";
    boolean   useTopic    = true;

    /*-----------------------------------------------------------------------
     * Variables
     *----------------------------------------------------------------------*/
    Connection      connection  = null;
    Session         session     = null;
    MessageConsumer msgConsumer = null;
    MessageProducer msgProducer = null;
    Message			replyMsg	= null;
    Destination     destination = null;
    Destination     replyDestination = null;
    boolean 		initialized = false;


    public tibjmsAsyncMsgConsumerRecover(String[] args)
    {
        parseArgs(args);

        /*
        try
        {
            tibjmsUtilities.initSSLParams(serverUrl,args);
        }
        catch(JMSSecurityException e)
        {
            System.err.println("JMSSecurityException: "+e.getMessage()+", provider="+e.getErrorCode());
            e.printStackTrace();
            System.exit(0);
        }
        */

        /* print parameters */
        System.err.println("------------------------------------------------------------------------");
        System.err.println("tibjmsAsyncMsgConsumerRecover Sample");
        System.err.println("------------------------------------------------------------------------");
        System.err.println("Server....................... " + (serverUrl != null ? serverUrl : "localhost"));
        System.err.println("User......................... " + (userName != null ? userName : "(null)"));
        System.err.println("Destination.................. " + name);
        System.err.println("ReplyDestination............. " + replyName);
        System.err.println("------------------------------------------------------------------------");

        /* Catching CTRL-C */
        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() { 
        		if (connection != null) {
        			try {
        				connection.close();
        			} catch (JMSException e1) {
        				e1.printStackTrace();
        			}
        		}
        		System.err.println("\nShutting down! CTRL-C");

            }
         });

        /* looping forever - same logic for initial connect and reconnect */
        do {
			try {
		        if (!initialized) 
		        	establishConnection();
				/* Sleep before checking if reconnect is needed */
		        Thread.sleep(1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
				break;
			}
        } while (true);
        
    }

    /*-----------------------------------------------------------------------
     * usage
     *----------------------------------------------------------------------*/
    void usage()
    {
        System.err.println("\nUsage: java tibjmsAsyncMsgConsumerRecover [options] [ssl options]");
        System.err.println("\n");
        System.err.println("   where options are:\n");
        System.err.println("\n");
        System.err.println(" -server   <server URL> - EMS server URL, default is local server\n");
        System.err.println(" -user     <user name>  - user name, default is null\n");
        System.err.println(" -password <password>   - password, default is null\n");
        System.err.println(" -topic    <topic-name> - topic name, default is \"topic.sample\"\n");
        System.err.println(" -queue    <queue-name> - queue name, no default\n");
        System.err.println(" -rtopic   <topic-name> - reply topic name, default is \"sample\"\n");
        System.err.println(" -rqueue   <queue-name> - reply queue name, no default\n");
        System.err.println(" -help-ssl              - help on ssl parameters\n");
        System.exit(0);
    }

    /*-----------------------------------------------------------------------
     * parseArgs
     *----------------------------------------------------------------------*/
    void parseArgs(String[] args)
    {
        int i=0;

        while(i < args.length)
        {
            if (args[i].compareTo("-server")==0)
            {
                if ((i+1) >= args.length) usage();
                serverUrl = args[i+1];
                i += 2;
            }
            else
            if (args[i].compareTo("-topic")==0)
            {
                if ((i+1) >= args.length) usage();
                name = args[i+1];
                i += 2;
            }
            else
            if (args[i].compareTo("-queue")==0)
            {
                if ((i+1) >= args.length) usage();
                name = args[i+1];
                i += 2;
                useTopic = false;
            }
            else
            if (args[i].compareTo("-rtopic")==0)
            {
                if ((i+1) >= args.length) usage();
                replyName = args[i+1];
                i += 2;
            }
            else
            if (args[i].compareTo("-rqueue")==0)
            {
                if ((i+1) >= args.length) usage();
                replyName = args[i+1];
                i += 2;
                useTopic = false;
            }
            else
            if (args[i].compareTo("-user")==0)
            {
                if ((i+1) >= args.length) usage();
                userName = args[i+1];
                i += 2;
            }
            else
            if (args[i].compareTo("-password")==0)
            {
                if ((i+1) >= args.length) usage();
                password = args[i+1];
                i += 2;
            }
            else
            if (args[i].compareTo("-help")==0)
            {
                usage();
            }
/*            else
            if (args[i].compareTo("-help-ssl")==0)
            {
                tibjmsUtilities.sslUsage();
            }
            else
            if(args[i].startsWith("-ssl"))
            {
                i += 2;
            }
*/
            else
            {
                System.err.println("Unrecognized parameter: "+args[i]);
                usage();
            }
        }
    }


    /*---------------------------------------------------------------------
     * onException
     *---------------------------------------------------------------------*/
    public void onException(JMSException e)
    {
        /* print the connection exception status */
        System.err.println("OnException: CONNECTION EXCEPTION: "+ e.getMessage());

		if (connection != null) {
			try {
				connection.close();
			} catch (JMSException e1) {
				e1.printStackTrace();
			}
		}
        /* Reconnect needed */
		initialized = false;
    }
    
    
    /*---------------------------------------------------------------------
     * establishConnection
     *---------------------------------------------------------------------*/
    private void establishConnection() {
		/* Stay in this loop until connect succeeds */
    	while(!initialized) {
			try
	        {
	            ConnectionFactory factory = new com.tibco.tibjms.TibjmsConnectionFactory(serverUrl);
	            /* set heartbeat interval to detect network issues */
	            Tibjms.setPingInterval(10);

	            /* create the connection */
	            connection = factory.createConnection(userName,password);

	            /* create the session */
	            session = connection.createSession(false, Session.CLIENT_ACKNOWLEDGE);
	            
	            /*
	            * create session in this mode if you want to use Tibjms.recover(msg); 
	            * session = connection.createSession(false, Tibjms.EXPLICIT_CLIENT_ACKNOWLEDGE);
	            */

	            /* set the exception listener */
	            connection.setExceptionListener(this);

	            /* create the destination */
	            if(useTopic) {
	                destination = session.createTopic(name);
	                replyDestination = session.createTopic(replyName);
	            }
	            else {
	                destination = session.createQueue(name);
	                replyDestination = session.createQueue(replyName);
	            }

	            System.err.println("Consumer to destination: "+name);
	            System.err.println("Producer to destination: "+replyName);


	            /* create the consumer */
	            msgConsumer = session.createConsumer(destination);

	            /* set the message listener */
	            msgConsumer.setMessageListener(this);

	            /* create the producer */
	            msgProducer = session.createProducer(null);

	            /* create text message */
                replyMsg = session.createTextMessage();

                /* set message text */
                ((TextMessage) replyMsg).setText("REPLY MESSAGE");


	            /* start the connection */
	            connection.start();

	            /* Connect succeeded */
	            initialized = true;
	        }
	        catch(Exception e)
	        {
	        	System.err.println("CONNECTION EXCEPTION: "+ e.getMessage());
	        	// e.printStackTrace();
	            try {
					/* Sleep before next connection attempt */
	            	Thread.sleep(1000);
				} catch (InterruptedException e1) {
					e1.printStackTrace();
				}
	        }

		}
		
	}


    // This is to simulate errors in message processing
    public Boolean processMessage(Message msg) {
		return (System.nanoTime() & 1) == 0;
    }
    
	/*---------------------------------------------------------------------
     * onMessage
     *---------------------------------------------------------------------*/
    public void onMessage(Message msg)
    {
        try
        {
            System.err.println("Received message: " + msg);
            if (processMessage(msg)) {
            	/* Message processing OK */
                /* Send reply message and acknowledge message */
            	msgProducer.send(msg.getJMSReplyTo(),replyMsg);
                msg.acknowledge();
                System.err.println("Sent message: " + replyMsg);
            } else {
            	/* Message processing failed due to external reason, 
            	 * e.g. jdbc write to database failed - retry needed
            	 * recover session, message re-delivery will follow and onMessage will be called again
            	 * redelivers 5 messages from server to client if prefetch=5
            	 */
            	session.recover();
            	/* 
            	* Tibjms.recoverMsg(msg); 	// recovers only this message, saves bandwidth - but order is not preserved
            	* 							//and remember to use the EXPLICIT_CLIENT_ACKNOWLEDGE mode
            	*/
            	Thread.sleep(15000); // Sleep 15 s, re-delivery will happen when onMessage ends
            }
        }
        catch(Exception e)
        {
            System.err.println("OnMessage: JMS Exception: " + e.getMessage());
        }
    }

    /*-----------------------------------------------------------------------
     * main
     *----------------------------------------------------------------------*/
    public static void main(String[] args)
    {
        new tibjmsAsyncMsgConsumerRecover(args);
    }

}


